functor PPrint(LambdaAst : LAMBDA_AST) : PPRINT = struct
  
  structure LambdaAst = LambdaAst

  val ppAlpha = ref false

  local 
    open LambdaAst Compiler.PrettyPrint
  in
    fun ppVar pps (s, i) =
      let val str = s ^ (if !ppAlpha then Int.toString i else "") in
	add_string pps str
      end

    fun ppNum pps i = 
      if i >= 0 then add_string pps (Int.toString i)
      else add_string pps ("-"^(Int.toString (~1 * i)))

    fun ppList printer pps style list =
      let 
	val pp = add_string pps
	fun ppListH [] = ()
	  | ppListH [x] = printer pps x
	  | ppListH (x::rest) = 
	  (printer pps x;
	   add_break pps (1, 0);
	   ppListH rest)
      in
	(begin_block pps style 0;
	 ppListH list;
	 end_block pps)
      end

    fun ppPrimop pps p = 
      let val pp = add_string pps in
      case p of 
	Plus =>  pp "+"
      | Times => pp "*"
      | Minus => pp "-"
      | ZeroP => pp "zero?"
      | Fst => pp "fst"
      | Snd => pp "snd"
      end

    fun ppExpr pps e = 
      let val pp = add_string pps in
      case e of
	Var v => ppVar pps v
      | Num i => ppNum pps i
      | Op p => ppPrimop pps p
      | Lambda(vl, e) => 
	  (pp "(fn ";
	   begin_block pps INCONSISTENT 2;
	   pp "(";
	   ppList ppVar pps INCONSISTENT vl;
	   pp ")";
	   add_break pps (1, 2);
	   ppExpr pps e;
	   end_block pps;
	   pp ")"
	   )
      | App el => 
	  (begin_block pps INCONSISTENT 0;
	   pp "(";
	   ppList ppExpr pps INCONSISTENT el;
	   pp ")";
	   end_block pps)
      | Pair (e1,e2) => 
	  (begin_block pps INCONSISTENT 0;
	   pp "<";
	   ppExpr pps e1;
	   pp ",";
	   ppExpr pps e2;
	   pp ">";
	   end_block pps)
      | Letrec (vel, e) => 
	  (pp "(letrec ";
	   begin_block pps INCONSISTENT 2;
	   pp "(";
	   ppList (fn pps => fn (v, e) => 
		   let val pp = add_string pps in
		   (pp "(";
		    ppVar pps v;
		    add_break pps (1, 0);
		    ppExpr pps e;
		    pp ")")
		   end)  pps CONSISTENT vel;
	   pp ")";
	   add_break pps (1, 2);
	   ppExpr pps e;
	   end_block pps;
	   pp ")")
      | Case(e,(v1,e1),(v2,e2)) => 
	  (pp "(case ";
	   begin_block pps INCONSISTENT 2;
	   ppExpr pps e;
	   (pp "(";
	    ppVar pps v1;
	    add_break pps (1, 0);
	    ppExpr pps e1;
	    pp ")");
	   (pp "(";
	    ppVar pps v2;
	    add_break pps (1, 0);
	    ppExpr pps e2;
	    pp ")");
	   end_block pps;
	   pp ")")
      | Left e => 
	  (pp "(left ";
	   ppExpr pps e;
	   pp ")")
      | Right e => 
	  (pp "(right ";
	   ppExpr pps e;
	   pp ")")
      end

    fun ppE e = 
      let val s = pp_to_string 75 ppExpr e in print s end

  end


end


